@@ -4,13 +4,13 @@ test.getLatestCumulative.RAA <- function() {
     ## Test the getLatestCumulative function on RAA data
     ## First check one value from normal RAA data
     latest <- ChainLadder:::getLatestCumulative(RAA)
-    checkEquals(latest[1], RAA[1, ncol(RAA)])
+    checkEquals(latest[1], RAA[1, ncol(RAA)], checkNames=FALSE)
 
     ## Now try again after removing some early valuations
     RAA2 <- RAA
     RAA2[row(RAA) + col(RAA) <= 4] <- NA
     latest2 <- ChainLadder:::getLatestCumulative(RAA2)
-    checkEquals(latest2[1], RAA2[1, ncol(RAA2)])
+    checkEquals(latest2[1], RAA2[1, ncol(RAA2)], checkNames=FALSE)
 }
 
 test.getLatestCumulative.simple <- function() {
@@ -19,12 +19,12 @@ test.getLatestCumulative.simple <- function() {
     ## First, a traditional triangle
     simple.tri1 <- rbind(c(1,2), c(3, NA))
     latest <- ChainLadder:::getLatestCumulative(simple.tri1)
-    checkEquals(latest, c(2,3))
+    checkEquals(latest, c(2,3), checkNames=FALSE)
 
     ## Now check very simple "trapezoidal" triangle
     simple.tri2 <- rbind(c(NA ,2), c(3, NA))
     latest2 <- ChainLadder:::getLatestCumulative(simple.tri2)
-    checkEquals(latest2, c(2,3))
+    checkEquals(latest2, c(2,3), checkNames=FALSE)
 }
 
 test.tailfactor <- function () {
